@@ -66,6 +66,7 @@
 #define AUTH_LOGIN      1 << 1
 #define AUTH_CRAM_MD5   1 << 2
 #define AUTH_DIGEST_MD5 1 << 3
+#define NUM_AUTH        4        /* number of auth methods to try */
 
 #define DIGEST_MD5_REALM_LEN   256
 #define DIGEST_MD5_NONCE_LEN   64
@@ -360,6 +361,7 @@ digest_md5(char *response, unsigned char *text, unsigned int text_len, const cha
 smtp_t *
 smtp_auth(config_t *cfg) {
 
+    int cnt;
     int s;
     struct sockaddr_in addr;
     struct hostent *he;
@@ -369,7 +371,9 @@ smtp_auth(config_t *cfg) {
     struct iovec iov[5];
     char *c;
     int rc;
+    int readBytes = 0;
     char rbuf[RESP_LEN];
+    char sockbuf[RESP_LEN];
     int auth = 0;
     int avail_auth_type = 0;
     char *tbuf;
@@ -386,6 +390,28 @@ smtp_auth(config_t *cfg) {
 #ifdef USE_SSL
     int use_ssl;
 #endif
+    // function pointer points to auth_xxxx();
+    int (*auth_method)(socket_t *, config_t *) = NULL;
+    // array that contains function pointer points to auth_xxxx();
+    int (*auth_methods[NUM_AUTH]) (socket_t *, config_t *) = { NULL };
+    int auth_methods_pos = 0;
+
+    char method_order[NUM_AUTH][16];
+    char *order_buf;
+
+    strcpy(method_order[0], strtok(global.trymechs," ,"));
+#ifdef DEBUG
+    log_debug(DEBUG_1, "smtp_auth: method_order[0] = %s",method_order[0]);
+#endif
+    for(cnt=1; ; cnt++) {
+        if(NULL == (order_buf = strtok(NULL, " ,"))) {
+            break;
+        }
+        strcpy(method_order[cnt], order_buf);
+#ifdef DEBUG
+        log_debug(DEBUG_1, "smtp_auth: method_order[%d] = %s",cnt,method_order[cnt]);
+#endif
+    }
 
     if(!cfg->password) {
         if(!global.password) {
@@ -465,9 +491,14 @@ smtp_auth(config_t *cfg) {
             }
         }
     }
-    addr.sin_addr.s_addr = *((int *)he->h_addr_list[0]);
     s = socket(PF_INET, SOCK_STREAM, 0);
 
+    addr.sin_family = AF_INET;
+    addr.sin_port = htons(cfg->port);
+    addr.sin_addr.s_addr = *((int *)he->h_addr_list[0]);
+    addr.sin_addr = *((struct in_addr *)he->h_addr);
+    memset(&(addr.sin_zero),'\0',8);
+
     if(cfg->conn_timeout > 0) {
         set_timeout(cfg->conn_timeout);
     }
@@ -505,7 +536,15 @@ smtp_auth(config_t *cfg) {
     if(cfg->timeout > 0) {
         set_timeout(cfg->timeout);
     }
-    rc = socket_read(smtp->sock, rbuf, sizeof(rbuf));
+
+    //////////////////////////////////
+    //usleep(3000000);
+    memset(rbuf,0x00,sizeof(rbuf));
+    memset(rbuf,0x00,sizeof(sockbuf));
+    rc = socket_read(smtp->sock, sockbuf, sizeof(rbuf)-1);
+#ifdef DEBUG
+    log_debug(DEBUG_1, "smtp_auth: read count: %d", rc);
+#endif
     if(cfg->timeout > 0) {
         set_timeout(0);
     }
@@ -519,11 +558,51 @@ smtp_auth(config_t *cfg) {
         strcpy(smtp->error_message, msgbuf);
         goto bail;
     }
-    rbuf[rc] = '\0';
+    readBytes += rc;
+#ifdef DEBUG
+    log_debug(DEBUG_1, "smtp_auth: 1st 220 response total read count %d", readBytes);
+#endif
+    sockbuf[rc+1] = '\0';
+    rbuf[readBytes+1] = '\0';
+    strcat(rbuf, sockbuf);
     c = strpbrk(rbuf, SMTP_NEWLINE);
-    if(c != NULL) {
-        *c = '\0';
+    if(c == NULL) {
+#ifdef DEBUG
+        log_debug(DEBUG_1, "smtp_auth: 220 response is not terminated. try to read again. now: %s", rbuf);
+#endif
+        rc = socket_read(smtp->sock, sockbuf, sizeof(sockbuf)-1);
+#ifdef DEBUG
+        log_debug(DEBUG_1, "smtp_auth: read count: %d", rc);
+#endif
+        if(cfg->timeout > 0) {
+		set_timeout(0);
+        }
+        if(rc == -1) {
+#ifdef DEBUG
+            log_debug(DEBUG_1, "smtp_auth: read (banner): %m");
+#endif
+            smtp->error = 1;
+            strcpy(msgbuf, RESP_SYNCERROR);
+            smtp->error_message = malloc(strlen(msgbuf) + 1);
+            strcpy(smtp->error_message, msgbuf);
+            goto bail;
+        }
+        readBytes += rc;
+#ifdef DEBUG
+        log_debug(DEBUG_1, "smtp_auth: 220 response total read count %d", readBytes);
+#endif
+        sockbuf[rc+1] = '\0';
+        rbuf[readBytes+1] = '\0';
+        strcat(rbuf, sockbuf);
     }
+    c = strpbrk(rbuf, SMTP_NEWLINE);
+    if (c != NULL) {
+#ifdef DEBUG
+        log_debug(DEBUG_1, "smtp_auth: 220 response is terminated. total read count: %d, now: %s", readBytes, rbuf);
+#endif
+        rbuf[readBytes+1] = '\0';
+    }
+    //////////////////////////////////
 
     if(strncmp(rbuf, "220 ", sizeof("220 ")-1)) {
 #ifdef DEBUG
@@ -575,14 +654,17 @@ smtp_auth(config_t *cfg) {
     if(cfg->timeout > 0) {
         set_timeout(cfg->timeout);
     }
+
+    memset(rbuf,0x00,sizeof(rbuf));
     rc = socket_read(smtp->sock, rbuf, sizeof(rbuf));
     if(cfg->timeout > 0) {
         set_timeout(0);
     }
-    if(rc == -1) {
 #ifdef DEBUG
-        log_debug(DEBUG_1, "smtp_auth: read (response): %m");
+    log_debug(DEBUG_1, "smtp_auth: read (response): %m");
+    log_debug(DEBUG_1, "smtp_auth: read (response): %s", rbuf);
 #endif
+    if(rc == -1) {
         smtp->error = 1;
         strcpy(msgbuf, RESP_IERROR);
         smtp->error_message = malloc(strlen(msgbuf) + 1);
@@ -630,19 +712,38 @@ smtp_auth(config_t *cfg) {
 #endif
 
     /* build the AUTH command */
-    if(avail_auth_type & AUTH_CRAM_MD5) {
-        auth = auth_cram_md5(smtp->sock,&global);
+    for(cnt=0; cnt<NUM_AUTH; cnt++) {
+        if (method_order[cnt] == NULL) {
+            continue;
+        }
+
+        if (strcmp(method_order[cnt],"CRAM-MD5")==0 && (avail_auth_type & AUTH_CRAM_MD5)) {
+            auth_methods[auth_methods_pos++] = auth_cram_md5;
+#ifdef DEBUG
+            log_debug(DEBUG_5, "set auth_methods[%d] = auth_cram_md5;", auth_methods_pos-1);
+#endif
+        }
+        if (strcmp(method_order[cnt],"LOGIN")==0 && (avail_auth_type & AUTH_LOGIN)) {
+            auth_methods[auth_methods_pos++] = auth_login;
+#ifdef DEBUG
+            log_debug(DEBUG_5, "set auth_methods[%d] = auth_login;", auth_methods_pos-1);
+#endif
+        }
+        if (strcmp(method_order[cnt],"PLAIN")==0 && (avail_auth_type & AUTH_PLAIN)) {
+            auth_methods[auth_methods_pos++] = auth_plain;
+#ifdef DEBUG
+            log_debug(DEBUG_5, "set auth_methods[%d] = auth_plain;", auth_methods_pos-1);
+#endif
+        }
+        if (strcmp(method_order[cnt],"DIGEST-MD5")==0 && (avail_auth_type & AUTH_DIGEST_MD5)) {
+            auth_methods[auth_methods_pos++] = auth_digest_md5;
+#ifdef DEBUG
+            log_debug(DEBUG_5, "set auth_methods[%d] = auth_digest_md5;", auth_methods_pos-1);
+#endif
+        }
     }
-    else if((avail_auth_type & AUTH_LOGIN) != 0) {
-        auth = auth_login(smtp->sock,&global);
-    }
-    else if((avail_auth_type & AUTH_PLAIN) != 0) {
-        auth = auth_plain(smtp->sock,&global);
-    }
-    else if((avail_auth_type & AUTH_DIGEST_MD5) != 0) {
-        auth = auth_digest_md5(smtp->sock,&global);
-    }
-    else {
+
+    if (auth_methods_pos == 0) {
 #ifdef DEBUG
         log_debug(DEBUG_1, "smtp_auth: smtp authentication is not implemented: %s", rbuf);
 #endif
@@ -653,6 +754,28 @@ smtp_auth(config_t *cfg) {
         goto bail;
     }
 
+    for (auth_methods_pos = 0; auth_methods_pos < NUM_AUTH; auth_methods_pos++) {
+      if (auth_methods[auth_methods_pos] == NULL) {
+#ifdef DEBUG
+        log_debug(DEBUG_1, "auth_methods[0] == NULL");
+#endif
+        break;
+      }
+      auth_method = auth_methods[auth_methods_pos];
+      // try to auth
+      auth = auth_method(smtp->sock, &global);
+      if (auth) {
+#ifdef DEBUG
+        log_debug(DEBUG_1, "auth = %d", auth);
+#endif
+        break;
+      } else {
+#ifdef DEBUG
+        log_debug(DEBUG_1, "auth failed. try next... = %d", auth);
+#endif
+      }
+    }
+
 #ifdef DEBUG
     log_debug(DEBUG_5, "smtp_auth: auth=%d", auth);
 #endif
